/*
 * Copyright (c) 2015-2016, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/***** Includes *****/
#include <stdlib.h>
#include <xdc/std.h>
#include <xdc/runtime/System.h>

#include <ti/sysbios/BIOS.h>
#include <ti/sysbios/knl/Task.h>

/* Drivers */
#include <ti/drivers/rf/RF.h>
#include <ti/drivers/PIN.h>
#include <ti/drivers/crypto/CryptoCC26XX.h>


/* Board Header files */
#include "Board.h"
#include "smartrf_settings/smartrf_settings.h"

/* Pin driver handle */
static PIN_Handle ledPinHandle;
static PIN_State ledPinState;

/*
 * Application LED pin configuration table:
 *   - All LEDs board LEDs are off.
 */
PIN_Config pinTable[] =
{
    Board_PIN_LED1 | PIN_GPIO_OUTPUT_EN | PIN_GPIO_LOW | PIN_PUSHPULL | PIN_DRVSTR_MAX,
    PIN_TERMINATE
};


/***** Defines *****/
#define TX_TASK_STACK_SIZE 1024
#define TX_TASK_PRIORITY   2

/* Packet TX Configuration */
#define PAYLOAD_LENGTH      24
#define PACKET_INTERVAL     (uint32_t)(4000000*0.5f) /* Set packet interval to 500ms */



/***** Prototypes *****/
static void txTaskFunction(UArg arg0, UArg arg1);
static void encryptPacketECB(uint8_t *key);


/***** Variable declarations *****/
static Task_Params txTaskParams;
Task_Struct txTask;    /* not static so you can see in ROV */
static uint8_t txTaskStack[TX_TASK_STACK_SIZE];

static RF_Object rfObject;
static RF_Handle rfHandle;

uint32_t curtime;
static uint8_t rawPacket[PAYLOAD_LENGTH];
// Allocate enough space for encrypted packet
static uint8_t packet[PAYLOAD_LENGTH + AES_ECB_LENGTH];
// Key variables
static CryptoCC26XX_KeyLocation keyLocation = CRYPTOCC26XX_KEY_0;
static uint8_t cryptoKey[16] = { 0x2B, 0x7E, 0x15, 0x16, 0x28, 0xAE, 0xD2, 0xA6,
                    0xAB, 0xF7, 0x15, 0x88, 0x09, 0xCF, 0x4F, 0x3C };
//static uint16_t seqNumber;
static PIN_Handle pinHandle;

void encryptPacketECB(uint8_t * key)
{
        uint8_t clearTextBuffer[AES_ECB_LENGTH];
        CryptoCC26XX_Handle             handle;
        int32_t                         keyIndex;
        int32_t                         status;
        CryptoCC26XX_AESECB_Transaction trans;
        // Initialize Crypto driver
        CryptoCC26XX_init();
        // Attempt to open CryptoCC26XX.
        handle = CryptoCC26XX_open(Board_CRYPTO0, false, NULL);
        if (!handle) {
          System_abort("Crypto module could not be opened.");
        }
        keyIndex = CryptoCC26XX_allocateKey(handle, keyLocation, (const uint32_t *) key);
        if (keyIndex == CRYPTOCC26XX_STATUS_ERROR) {
                System_abort("Key Location was not allocated.");
        }

        CryptoCC26XX_Transac_init((CryptoCC26XX_Transaction *) &trans, CRYPTOCC26XX_OP_AES_ECB_ENCRYPT);
        // Setup transaction
        trans.keyIndex         = keyIndex;
        uint8_t i;
        for (i = 0; i < (sizeof(rawPacket)/AES_ECB_LENGTH); i++)
        {
           trans.msgIn            = (uint32_t *) &rawPacket[i * AES_ECB_LENGTH];
           trans.msgOut           = (uint32_t *) &packet[i*AES_ECB_LENGTH];
           // Encrypt the plaintext with AES ECB
           status = CryptoCC26XX_transact(handle, (CryptoCC26XX_Transaction *) &trans);
           if(status != CRYPTOCC26XX_STATUS_SUCCESS){
               System_abort("Encryption failed.");
           }
        }
        if (sizeof(packet)%AES_ECB_LENGTH != 0)
        {
            memset(clearTextBuffer, 0x0, AES_ECB_LENGTH);
            memcpy(clearTextBuffer, &rawPacket[i*AES_ECB_LENGTH],(i+1)*AES_ECB_LENGTH-sizeof(rawPacket));
            trans.msgIn            = (uint32_t *) clearTextBuffer;
            trans.msgOut           = (uint32_t *) &packet[i*AES_ECB_LENGTH];

            // Encrypt the plaintext with AES ECB
            status = CryptoCC26XX_transact(handle, (CryptoCC26XX_Transaction *) &trans);
            if(status != CRYPTOCC26XX_STATUS_SUCCESS){
               System_abort("Encryption failed.");
            }
        }
        CryptoCC26XX_releaseKey(handle, &keyIndex);
        CryptoCC26XX_close(handle);
}


/***** Function definitions *****/
void TxTask_init(PIN_Handle inPinHandle)
{
    pinHandle = inPinHandle;

    Task_Params_init(&txTaskParams);
    txTaskParams.stackSize = TX_TASK_STACK_SIZE;
    txTaskParams.priority = TX_TASK_PRIORITY;
    txTaskParams.stack = &txTaskStack;
    txTaskParams.arg0 = (UInt)1000000;

    Task_construct(&txTask, txTaskFunction, &txTaskParams, NULL);
}

static void txTaskFunction(UArg arg0, UArg arg1)
{
    uint32_t curtime;
    RF_Params rfParams;
    RF_Params_init(&rfParams);
    RF_cmdPropTx.pktLen = (PAYLOAD_LENGTH / AES_ECB_LENGTH + 1)*AES_ECB_LENGTH;
    //RF_cmdPropTx.pktLen = PAYLOAD_LENGTH;
    RF_cmdPropTx.pPkt = packet;
    RF_cmdPropTx.startTrigger.triggerType = TRIG_NOW;
    RF_cmdPropTx.startTrigger.pastTrig = 1;
    RF_cmdPropTx.startTime = 0;

    /* Request access to the radio */
    rfHandle = RF_open(&rfObject, &RF_prop, (RF_RadioSetup*)&RF_cmdPropRadioDivSetup, &rfParams);

    /* Set the frequency */
    RF_postCmd(rfHandle, (RF_Op*)&RF_cmdFs, RF_PriorityNormal, NULL, 0);

    /* Get current time */
    curtime = RF_getCurrentTime();

    while(1)
    {
        /* Create packet with incrementing sequence number and random payload */
        //rawPacket[0] = (uint8_t)(seqNumber >> 8);
        //rawPacket[1] = (uint8_t)(seqNumber++);

        rawPacket[0] = 0x00;
        rawPacket[1] = 0x01;

        uint8_t i;
        for (i = 2; i < PAYLOAD_LENGTH; i++)
        {
            //rawPacket[i] = rand();
            rawPacket[i] = i;
        }

        encryptPacketECB(cryptoKey);
        //decryptPacketECB(cryptoKey);

        curtime += PACKET_INTERVAL;
        RF_cmdPropTx.startTime = curtime;

        /* Send packet */
        RF_EventMask result = RF_runCmd(rfHandle, (RF_Op*)&RF_cmdPropTx, RF_PriorityNormal, NULL, 0);
        if (!(result & RF_EventLastCmdDone))
        {
            /* Error */
            while(1);
        }

//        RF_yield(rfHandle);

        PIN_setOutputValue(pinHandle, Board_PIN_LED1,!PIN_getOutputValue(Board_PIN_LED1));

        Task_sleep(5000000/Clock_tickPeriod);
    }
}

/*
 *  ======== main ========
 */
int main(void)
{
    /* Call driver init functions. */
    Board_initGeneral();

    /* Open LED pins */
    ledPinHandle = PIN_open(&ledPinState, pinTable);
    if(!ledPinHandle)
    {
        System_abort("Error initializing board LED pins\n");
    }

    /* Initialize task */
    TxTask_init(ledPinHandle);

    /* Start BIOS */
    BIOS_start();

    return (0);
}
